--Use lib stub to figure out if the same or higher version was already used
local major, minor = "LibConstantMapper", 0.6
local lib = LibStub:NewLibrary(major, minor)

if not lib then
	return
end

local function CompareKeyValuePair(a, b)
  --assert valid input
  assert(type(a) == "table" and a.key ~= nil and a.value ~= nil, "argument a must be a table containing a 'key' and 'value' string-index")
  assert(type(b) == "table" and b.key ~= nil and b.value ~= nil, "argument b must be a table containing a 'key' and 'value' string-index")
  
  --Don't try to compare data that is not comparable
  local typeValA, typeValB, ValNotComparable = type(a.value), type(b.value), false
  if(typeValA == "userdata" or typeValB == "userdata") then
      ValNotComparable = true
  end
  
  --check if the value is smaller first. Don't do that if values are not comparable via <
  if not ValNotComparable and (a.value < b.value) then
	return true
  end

  --if values are the same or non comparable, try to order by key
  if (ValNotComparable or (a.value == b.value)) and (a.key < b.key) then
	return true
  end

  return false
end

function lib:getValIndexedKeyList(input, noDuplicateAssertion)
	--assert that input is remotely valid
	assert (input ~= nil and type(input) == "table", "input must be not nil and a table")
	
	local result = {}

	for outerkey, outervalue in ipairs(input) do
		--assert that this index of input has the right contained elements
		assert (type(outervalue) == "table" and outervalue.key ~= nil and outervalue.value ~= nil, "Key: '" .. outerkey .. "' is not a table or does not contain a pair of key/value indexes")

		--Some local variables to shorten the syntax
		local innerkey, innervalue = outervalue.key, outervalue.value

		--Check if this value has already been used as key
		assert(noDuplicateAssertion or result[innervalue] == nil, "Key: '" .. outerkey .. "' contains the duplicate value '" .. innervalue .. "'")
		result[innervalue] = innerkey
	end

	return result
end

function lib:getMappedData(pattern, exclude)
	--Sanity check on values
	assert(pattern ~= nil and type(pattern) == "string", "pattern must not be nil and of type string")
	assert(exlcude == nil or type(exclude) == "string", "exclude must be nil or of type string")

	local result = {}

	--For every entry in the Global table
	for key, value in zo_insecurePairs(_G) do
		--Compare it to the given mapping
		if (key):find(pattern) and (exclude == nil or not(key):find(exclude)) then
			--found a Value for this mapping, so store it in the result table
			local newentry = { key = key, value = value}

			table.insert(result, newentry)
		end
	end

	return result
end

function lib:getDataByMapping(map_name, rebuildCache)
	local map = self:getMap(map_name)
	assert (map ~= nil and map ~= {}, "A mapping named '" .. map_name .. "' could not be found")

	--if there is nothing in the cache or ignoring the chache was ordered, build it from scratch
	if(self.__cache[map] == nil or rebuildCache) then
		d("Building Cache for map: " .. map_name)
		--If all worked out so far, let getMappedData do it's job
		local result = self:getMappedData(map.pattern, map.exclude)

		if result ~= {} then
			table.sort(result, CompareKeyValuePair)
		end

		self.__cache[map] = result
	end

	--get the cached value
	local result = self.__cache[map]

	--clone it
	result = ZO_DeepTableCopy(result)

	--return it
	return result
end

--Functions and Data related to the maps
function lib:listMaps()
	local result = {}

	for key, value in pairs(self.__maps) do
		table.insert(result, key)
	end

	return result
end

function lib:addMap(map, pattern, exclude)
	--create the map and cache tables, if they are missing
	self.__maps = self.__maps or {}
	self.__cache = self.__cache or {}

	assert(type(map) == "string", "map must not be nil and of type string")
	assert(type(pattern) == "string", "pattern must not be nil and of type string")
	assert(exlcude == nil or type(exclude) == "string", "exclude must be nil or of type string")

	--register the map. Reset the cache if there is any
	self.__maps[map] = { pattern = pattern, exclude = exclude }
	self.__cache[map] = nil
end

function lib:getMap(map)
	assert(type(map) == "string", "map must not be nil and of type string")

	return self.__maps[map]
end

--Predefined mappings
local ConstantMapperMappings = {
	{ mapping = "BindingTypes", pattern = "^BIND_TYPE_" },
	{ mapping = "ChatCategories", pattern = "^CHAT_CATEGORY_", exclude = "^CHAT_CATEGORY_HEADER" },
	{ mapping = "ChatChannels", pattern = "^CHAT_CHANNEL_"},
	{ mapping = "Events", pattern = "^EVENT_"},
	{ mapping = "EventsAbility", pattern = "^EVENT_ABILITY_" },
	{ mapping = "EventsAchievement", pattern = "^EVENT_ACHIEVEMENT_"},
	{ mapping = "EventsAction", pattern = "^EVENT_ACTION_"},
	{ mapping = "EventsActive", pattern = "^EVENT_ACTIVE_"},
	{ mapping = "EventsAddOn", pattern = "^EVENT_ADD_ON_" },
	{ mapping = "EventsAgentChat", pattern = "^EVENT_AGENT_CHAT_" },
	{ mapping = "EventsAlliancePoint", pattern = "^EVENT_ALLIANCE_POINT_" },
	{ mapping = "EventsArtifact", pattern = "^EVENT_ARTIFACT_" },
	{ mapping = "EventsAssignedCampaign", pattern = "^EVENT_ASSIGNED_CAMPAIGN_" },
	{ mapping = "EventsAttribute", pattern = "^EVENT_ATTRIBUTE_" },
	{ mapping = "EventsAvenge", pattern = "^EVENT_AVENGE_" },
	{ mapping = "EventsBankIs", pattern = "^EVENT_BANK_IS_" },
	{ mapping = "EventsBankedMoney", pattern = "^EVENT_BANKED_MONEY_" },
	{ mapping = "EventsBattleStandards", pattern = "^EVENT_BATTLE_STANDARDS_" },
	{ mapping = "EventsBegin", pattern = "^EVENT_BEGIN_" },
	{ mapping = "EventsBosses", pattern = "^EVENT_BOSSES_" },
	{ mapping = "EventsBroadcast", pattern = "^EVENT_BROADCAST" },
	{ mapping = "EventsBuy", pattern = "^EVENT_BUY_" },
	{ mapping = "EventsBuyback", pattern = "^EVENT_BUYBACK_" },
	{ mapping = "EventsCampaign", pattern = "^EVENT_CAMPAIGN_" },
	{ mapping = "EventsCancel", pattern = "^EVENT_CANCEL_" },
	{ mapping = "EventsCannot", pattern = "^EVENT_CANNOT_" },
	{ mapping = "EventsCapsLock", pattern = "^EVENT_CAPS_LOCK_STATE_" },
	{ mapping = "EventsCaptureArea", pattern = "^EVENT_CAPTURE_AREA_" },
	{ mapping = "EventsChatChannel", pattern = "^EVENT_CHAT_CHANNEL_" },
	{ mapping = "EventsChatLog", pattern = "^EVENT_CHAT_LOG_" },
	{ mapping = "EventsChatMessage", pattern = "^EVENT_CHAT_MESSAGE_" },
	{ mapping = "EventsChatter", pattern = "^EVENT_CHATTER_" },
	{ mapping = "EventsClose", pattern = "^EVENT_CLOSE_" },
	{ mapping = "EventsCombat", pattern = "^EVENT_COMBAT_" },
	{ mapping = "EventsConfirm", pattern = "^EVENT_CONFIRM_" },
	{ mapping = "EventsConversation", pattern = "^EVENT_CONVERSATION_" },
	{ mapping = "Zos", pattern = "^Zo" },
	{ mapping = "ZoFonts", pattern = "^ZoFont" }
}

for i=1, #ConstantMapperMappings, 1 do
	local currentMap = ConstantMapperMappings[i]
	lib:addMap(currentMap.mapping, currentMap.pattern, currentMap.exclude)
end

--set up slash commands and helper functions
local function slashHelper(input)
	d(lib:getDataByMapping(input))
end

SLASH_COMMANDS["/libcm"] = slashHelper